#!/usr/bin/env python


"""ZEd-Filterized Pub/sub Encapsulation Tool""" # "zProxy"


import array, select, socket, struct


# EthInterface.read -> Forwarder.handle_in/rid
#      -> ServerProxy.add_subscription, ClientProxy.send
# ServerProxy.read -> ... -> ServerProxy.publish -> Forwarder.handle_out/fid
#     -> EthInterface.send
# ClientProxy.subscribe -> Forwarder.handle_out/fid
#     -> EthInterface.send
#
# Additional features:
# - Unsubscriptions
# - Resubscriptions
# - Publish metadata about new streams
# - Correct headers


class PubSubPacket(object):
    ID_LEN = 32; FID_LEN = 32; MD_LEN = (2+2*ID_LEN+1*FID_LEN+3*8+1+5)
    INIT_OFFSET = 14
    TYPE_OFFSET = (INIT_OFFSET+1); SUB_TYPE = 2; PUB_TYPE = 4
    FID_OFFSET = (TYPE_OFFSET+4+5)
    RID_OFFSET = (FID_OFFSET+FID_LEN)+(2+2+1*ID_LEN) # XXX
    LEN_OFFSET = (RID_OFFSET+1*ID_LEN)
    PAYLOAD_OFFSET = (LEN_OFFSET+8+1+5)
    RFID_OFFSET = PAYLOAD_OFFSET+(2+2*ID_LEN)
    
    def __init__(self, packet):
        if not isinstance(packet, array.array):
            packet = array.array("c", packet)
        self.packet = packet
    
    @staticmethod
    def arror(dst, dst_offset, src, src_offset, length, copy_dst=False):
        if copy_dst:
            dst = array.array('c', dst[dst_offset:dst_offset+len])
            dst_offset = 0
        for i in xrange(length):
            dst_i = ord(dst[dst_offset+i]); src_i = ord(src[src_offset+i])
            dst_i |= src_i
#            print("%02x | %02x = %02x" \
#                      % (ord(dst[dst_offset+i]), dst_i, src_i))
            dst[dst_offset+i] = chr(dst_i)
        return dst

    @classmethod
    def hextobinid(cls, hexstr):
        import binascii
        
        if not (len(hexstr) % 2 == 0 \
                    and ("::" in hexstr or len(hexstr) == cls.FID_LEN*2)):
            raise ValueError("Bad Id string: bad length")
        
        substrings = hexstr.split("::")
        if len(substrings) > 2:
            raise ValueError("Bad Id string: \"::\" occurs more than once")
        for ss in substrings:
            if len(ss) % 2 != 0:
                raise ValueError("Bad Id string: contains single digits")
        
        if len(substrings) == 2:
            hexstr_ = ((cls.FID_LEN*2-len(hexstr)+2)*'0').join(substrings)
        else:
            hexstr_ = substrings[0]
        
        binid = binascii.unhexlify(hexstr_)
        assert len(binid) == cls.FID_LEN
        
        return binid


class ServerProxy(object):
    BUFSIZE = 1024
    
    def __init__(self, port, initial_fid, forwarder):
        self.port = port
        self.fid = initial_fid
        self.forwarder = forwarder
        
        # Create packet
        initial_packet = \
            (PubSubPacket.PAYLOAD_OFFSET + self.BUFSIZE) * "\x00"
        self.ps_packet = PubSubPacket(initial_packet)
        # Set type
        struct.pack_into("B", self.ps_packet.packet, self.ps_packet.TYPE_OFFSET,
                         self.ps_packet.PUB_TYPE)
        # Set FId
        fid_i1 = self.ps_packet.FID_OFFSET
        fid_i2 = fid_i1 + self.ps_packet.FID_LEN
        self.ps_packet.packet[fid_i1:fid_i2] = array.array("c", self.fid)
        # Set RId (port)
        struct.pack_into("I", self.ps_packet.packet, self.ps_packet.RID_OFFSET,
                         self.port) # XXX
        
        # Create socket
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.udp_socket.bind(("", self.port))
    
    def fileno(self):
        return self.udp_socket.fileno()
    
    def _read(self):
        return array.array("c", self.udp_socket.recv(self.BUFSIZE))
    
    def receive(self):
        return self.publish(self._read())
    
    def publish(self, data):
        data_len = len(data)
        struct.pack_into("I", self.ps_packet.packet, self.ps_packet.LEN_OFFSET,
                         data_len)
#        payload_i1 = self.ps_packet.PAYLOAD_OFFSET
#        payload_i2 = payload_i1 + len(data)
#        self.ps_packet.packet[payload_i1:payload_i2] = data
        self.ps_packet.packet[self.ps_packet.PAYLOAD_OFFSET:] = data
        
        return self.forwarder.handle_out(self.ps_packet)
    
    def add_subscription(self, ps_packet):
        type = \
            struct.unpack_from("B", ps_packet.packet, ps_packet.TYPE_OFFSET)[0]
        if type == ps_packet.SUB_TYPE:
            self.ps_packet.arror(self.ps_packet.packet,
                                 self.ps_packet.FID_OFFSET,
                                 ps_packet.packet,
                                 ps_packet.RFID_OFFSET,
                                 self.ps_packet.FID_LEN)
#            print("Packet template: %r" % self.ps_packet.packet)
        # Currently we don't support unsubscribing, but we could do that
        # if we stored all added FIds for re-creating the ORed FId.



class ClientProxy(object):
    
    def __init__(self, port, sub_fid, forwarder):
        self.address = "localhost"
        self.port = port
        self.sub_fid = sub_fid
        self.forwarder = forwarder
        
        # Create socket
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    
    def send(self, ps_packet):
        type = \
            struct.unpack_from("B", ps_packet.packet, ps_packet.TYPE_OFFSET)[0]
        if type == ps_packet.PUB_TYPE:
            data_len = \
                struct.unpack_from("I",
                                   ps_packet.packet, ps_packet.LEN_OFFSET)[0]
            payload_i1 = ps_packet.PAYLOAD_OFFSET
            payload_i2 = payload_i1 + data_len
            data = ps_packet.packet[payload_i1:payload_i2]
            return self.udp_socket.sendto(
                buffer(data), (self.address, self.port))
    
    def subscribe(self):
        # Create packet
        initial_packet = \
            (PubSubPacket.PAYLOAD_OFFSET + PubSubPacket.MD_LEN) * "\x00"
        ps_packet = PubSubPacket(initial_packet)
        # Set type
        struct.pack_into("B", ps_packet.packet, ps_packet.TYPE_OFFSET,
                         ps_packet.SUB_TYPE)
        # Set FId
        fid_i1 = ps_packet.FID_OFFSET
        fid_i2 = fid_i1 + ps_packet.FID_LEN
        ps_packet.packet[fid_i1:fid_i2] = array.array("c", self.sub_fid)
        # Set RId (port)
        struct.pack_into("I", ps_packet.packet, ps_packet.RID_OFFSET,
                         self.port) # XXX
        
        self.forwarder.handle_out(ps_packet)
    

class EthInterface(object):
    BUFSIZE = 1514
    BC_ADDR = 6*"\xff"
    
    def __init__(self, interface, lid, proto, forwarder):
        self.interface = interface
        self.lid = lid
        self.proto = proto
        self.forwarder = forwarder
        
        self.eth_socket = \
            socket.socket(socket.AF_PACKET, socket.SOCK_RAW, proto)
        self.eth_socket.bind((interface, proto))
        
        if_name, if_proto, pkt_type, hw_type, hw_addr = \
            self.eth_socket.getsockname()
        self.hw_addr = hw_addr
    
    def fileno(self):
        return self.eth_socket.fileno()
        
    def send(self, ps_packet):
        # Set Ethernet II header
        struct.pack_into("!6s6sH", ps_packet.packet, 0,
                         self.BC_ADDR, self.hw_addr, self.proto)
        
#        print("Sending out %d bytes on interface %s" \
#                  % (len(ps_packet.packet), self.interface))
        return self.eth_socket.sendall(buffer(ps_packet.packet))
    
    def _read(self):
        return array.array("c", self.eth_socket.recv(self.BUFSIZE))
    
    def receive(self):
        ps_packet = PubSubPacket(self._read())
        # Add LId to packet
        ps_packet.arror(ps_packet.packet, ps_packet.RFID_OFFSET,
                        self.lid, 0,
                        ps_packet.FID_LEN);
        return self.forwarder.handle_in(ps_packet)
    

class Forwarder(object):
    def __init__(self):
        self.in_handlers = {} # {port/RId: [handler, ...], ...}
        self.out_handlers = {}  # {LId: handler, ...}
    
    def add_in_handler(self, port, handler):
        if port not in self.in_handlers:
            self.in_handlers[port] = set([])
        self.in_handlers[port].add(handler)
    
    def set_out_handler(self, lid, handler):
        self.out_handlers[lid] = handler
    
    def handle_in(self, ps_packet):
        port = \
            struct.unpack_from("I", ps_packet.packet, ps_packet.RID_OFFSET)[0]
        print("handle_in: %d" % port)
        if port in self.in_handlers:
            for handler in self.in_handlers[port]:
                handler(ps_packet)
    
    def handle_out(self, ps_packet):
        for lid, handler in self.out_handlers.items():
            if self.fidand(ps_packet.packet, ps_packet.FID_OFFSET,
                           lid, 0,
                           len(lid)):
                handler(ps_packet)
    
    @staticmethod
    def fidand(fid, fid_offset, lid, lid_offset, length):
        i = fid_offset+1; j = lid_offset+1
        for k in xrange(1, length):
            fid_i = ord(fid[i]); lid_j = ord(lid[j])
            if fid_i & lid_j != lid_j:
#                print("%02x & %02x != %02x" % (fid_i, lid_j, lid_j))
                return False
            i += 1; j += 1
        return True

def _main():
    import getopt, sys
        
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "s:c:f:g:i:e:h", ["help"])
        #print(opts, args)
    except getopt.GetoptError, err:
        print(str(err))
        sys.exit(1)
    
    server_proxies = set([])
    client_proxies = set([])
    default_pub_fid = PubSubPacket.FID_LEN*"\x00"
    default_sub_fid = PubSubPacket.FID_LEN*"\x00"
    interfaces = set([])
    proto = 0xacdc
    forwarder = Forwarder()
    
    # Scan for help
    for o, a in opts:
        if o == "-h" or o == "--help":
            print("%s %s" % (str(__file__).split("/")[-1],
                             "\n   -s server_port1[,pub_fid1] ..."
                             "\n   -c client_port1[,sub_fid1] ..."
                             "\n   -f default_pub_fid"
                             "\n   -g default_sub_fid"
                             "\n   -i interface1,lid1"
                             "\n   -e ethertype"
                             "\n   -h (help)"))
            sys.exit()
    
    # Scan for default FIds
    for o, a in opts:
        if o == "-f":
            default_pub_fid = PubSubPacket.hextobinid(a)
        if o == "-g":
            default_sub_fid = PubSubPacket.hextobinid(a)
    
    # Scan for Ethertype
    for o, a in opts:
        if o == "-e":
            proto = int(a, 16)
    
    # Scan for ports and interfaces
    for o, a in opts:
        if o in ("-f", "-g", "-e"):
            pass
        elif o == "-s":
            al = a.split(",")
            port = int(al[0])
            if (len(al) >= 2):
                pub_fid = PubSubPacket.hextobinid(al[1]) 
            else:
                pub_fid = default_pub_fid
            server_proxy = ServerProxy(port, pub_fid, forwarder)
            server_proxies.add(server_proxy)
            forwarder.add_in_handler(port, server_proxy.add_subscription)
        elif o == "-c":
            al = a.split(",")
            port = int(al[0])
            if (len(al) >= 2):
                sub_fid = PubSubPacket.hextobinid(al[1]) 
            else:
                sub_fid = default_sub_fid
            client_proxy = ClientProxy(port, sub_fid, forwarder)
            client_proxies.add(client_proxy)
            forwarder.add_in_handler(port, client_proxy.send)
        elif o == "-i":
            al = a.split(",")
            interface = al[0]
            lid = PubSubPacket.hextobinid(al[1])
            eth_interface = EthInterface(interface, lid, proto, forwarder) 
            interfaces.add(eth_interface)
            forwarder.set_out_handler(lid, eth_interface.send)
        else:
            assert False, "Unhandled option: %r" % o
    
    for client_proxy in client_proxies:
        client_proxy.subscribe()        
    
    print("Enter select")
    recv_set = server_proxies.union(interfaces)
    if len(recv_set) == 0:
        print("Warning: input list is empty")
    while True:
        l = select.select(recv_set, (), ())
        if len(l[0]) > 0:
            l[0][0].receive()
    

if __name__ == "__main__":
    _main()
