#!/usr/bin/env python


"""ZEd-Filterized Pub/sub Encapsulation Tool""" # "zProxy"


import array, binascii, select, signal, socket, struct, sys, time


# EthInterface.read -> Forwarder.handle_in/rid
#      -> ServerProxy.add_subscription, ClientProxy.send
# ServerProxy.read -> ... -> ServerProxy.publish -> Forwarder.handle_out/fid
#     -> EthInterface.send
# ClientProxy.subscribe -> Forwarder.handle_out/fid
#     -> EthInterface.send
#
# Additional features:
# - Resubscriptions
# - Publish metadata about new streams
# - Correct headers


class PubSubPacket(object):
    ID_LEN = 32; FID_LEN = 32; MD_LEN = (2+2*ID_LEN+1*FID_LEN+3*8+1+5)
    INIT_OFFSET = 14
    TYPE_OFFSET = (INIT_OFFSET+1); SUB_TYPE = 2; PUB_TYPE = 4; UNSUB_TYPE = 3
    TTL_OFFSET = (TYPE_OFFSET+1)
    FID_OFFSET = (TTL_OFFSET+3+5)
    RID_OFFSET = (FID_OFFSET+FID_LEN)+(2+1*ID_LEN) # XXX
    LEN_OFFSET = (RID_OFFSET+1*ID_LEN)
    PAYLOAD_OFFSET = (LEN_OFFSET+8+1+5)
    RFID_OFFSET = PAYLOAD_OFFSET+(2+2*ID_LEN)
    
    def __init__(self, packet=PAYLOAD_OFFSET*"\x00"):
        if not isinstance(packet, array.array):
            packet_ = array.array("c", packet)
        else:
            packet_ = packet
        self.packet = packet_
        
#        self.type = rwbuffer(self.packet,
#                             self.TYPE_OFFSET, 1)
#        self.ttl = rwbuffer(self.packet,
#                            self.TTL_OFFSET, 1)
        self.fid = buffer(self.packet,
                          self.FID_OFFSET, self.FID_LEN)
#        self.rid = rwbuffer(self.packet,
#                            self.RID_OFFSET, self.ID_LEN)
#        self.len = rwbuffer(self.packet,
#                            self.LEN_OFFSET, 8)
#        self.payload = rwbuffer(self.packet,
#                                self.PAYLOAD_OFFSET,
#                                len(self.packet))
        self.rfid = buffer(self.packet,
                           self.RFID_OFFSET, self.FID_LEN)
        
        self.packet_len = len(self.packet)
        self.packet_type = 0
#        self.packet_addr = self.packet.buffer_info()[0]
    
    @staticmethod
    def arror(dst, dst_offset, src, src_offset, length, copy_dst=False):
        if copy_dst:
            dst = array.array('c', dst[dst_offset:dst_offset+len])
            dst_offset = 0
        for i in xrange(length):
            tmp = ord(dst[dst_offset+i])
            tmp |= ord(src[src_offset+i])
            dst[dst_offset+i] = chr(tmp)
        return dst

    @classmethod
    def hextobinid(cls, hexstr):
        if not (len(hexstr) % 2 == 0 \
                    and ("::" in hexstr or len(hexstr) == cls.FID_LEN*2)):
            raise ValueError("Bad Id string: bad length")
        
        substrings = hexstr.split("::")
        if len(substrings) > 2:
            raise ValueError("Bad Id string: \"::\" occurs more than once")
        for ss in substrings:
            if len(ss) % 2 != 0:
                raise ValueError("Bad Id string: contains single digits")
        
        if len(substrings) == 2:
            hexstr_ = ((cls.FID_LEN*2-len(hexstr)+2)*'0').join(substrings)
        else:
            hexstr_ = substrings[0]
        
        binid = binascii.unhexlify(hexstr_)
        assert len(binid) == cls.FID_LEN
        
        return binid
    
    @staticmethod
    def bintohexid(binid):
        return binascii.hexlify(binid)


class ServerProxy(object):
    BUFSIZE = 1024
    
    def __init__(self, port, initial_fid, forwarder):
        self.port = port
        self.initial_fid = initial_fid
        self.forwarder = forwarder
        
        # Create packet
        initial_packet = \
            (PubSubPacket.PAYLOAD_OFFSET+self.BUFSIZE)*"\x00"
        self.ps_packet = PubSubPacket(initial_packet)
        
        # Set type
#        struct.pack_into("B", self.ps_packet.type, 0, PubSubPacket.PUB_TYPE)
        struct.pack_into("B", self.ps_packet.packet, PubSubPacket.TYPE_OFFSET,
                         PubSubPacket.PUB_TYPE)
        # Set TTL
        self.ps_packet.packet[PubSubPacket.TTL_OFFSET] = "\xff"
        # Set FId
#        self.ps_packet.fid[:] = self.initial_fid
        fid_i1 = PubSubPacket.FID_OFFSET
        fid_i2 = fid_i1 + PubSubPacket.FID_LEN
        self.ps_packet.packet[fid_i1:fid_i2] = \
            array.array("c", self.initial_fid)
        # Set RId (port)
#        struct.pack_into("I", self.ps_packet.rid, 0, self.port) # XXX
        struct.pack_into("I", self.ps_packet.packet, PubSubPacket.RID_OFFSET,
                         self.port) # XXX
        
        # Create socket
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.udp_socket.bind(("", self.port))
        
        self.recvbuf = array.array("c", self.BUFSIZE*"\x00")
        
        self.subscriptions = {}
    
    def fileno(self):
        return self.udp_socket.fileno()
    
    def _read(self):
#        return self.udp_socket.recv_into(self.ps_packet.payload, self.BUFSIZE)
        data_len = self.udp_socket.recv_into(self.recvbuf, self.BUFSIZE)
        payload_i1 = self.ps_packet.PAYLOAD_OFFSET
        payload_i2 = payload_i1 + data_len
        self.ps_packet.packet[payload_i1:payload_i2] = self.recvbuf[:data_len]
        return data_len
    
    def receive(self):
        return self.publish(self._read())
    
    def publish(self, data_len):
 #       struct.pack_into("I", self.ps_packet.len, 0, data_len)
        struct.pack_into("I", self.ps_packet.packet, PubSubPacket.LEN_OFFSET,
                         data_len)
        self.ps_packet.packet_len = PubSubPacket.PAYLOAD_OFFSET + data_len
        return self.forwarder.handle_out(self.ps_packet)
    
    def update_subscription(self, ps_packet):
        if ps_packet.packet_type == PubSubPacket.SUB_TYPE:
            print("Add subscription to %d:\n  %s" \
                      % (self.port, PubSubPacket.bintohexid(ps_packet.rfid)))
            
            # Or FIds
#            fidor(self.ps_packet.packet_addr, PubSubPacket.FID_OFFSET,
#                  ps_packet.packet_addr, PubSubPacket.RFID_OFFSET,
#                  PubSubPacket.FID_LEN)
            PubSubPacket.arror(self.ps_packet.packet,
                               PubSubPacket.FID_OFFSET,
                               ps_packet.packet,
                               PubSubPacket.RFID_OFFSET,
                               PubSubPacket.FID_LEN)
            
#            rfid = ps_packet.rfid[:]
            rfid_i1 = PubSubPacket.RFID_OFFSET
            rfid_i2 = rfid_i1 + PubSubPacket.FID_LEN
            rfid = ps_packet.packet[rfid_i1:rfid_i2].tostring()
            if rfid not in self.subscriptions:
                rfid_arr = array.array("c", rfid)
#                rfid_addr = rfid_arr.buffer_info()[0]
                self.subscriptions[rfid] = [1, rfid_arr] #, rfid_addr]
            else:
                self.subscriptions[ps_packet.rfid][0] += 1
            
            print("FId:\n  %s" \
                  % PubSubPacket.bintohexid(self.ps_packet.fid))
        
        elif ps_packet.packet_type == PubSubPacket.UNSUB_TYPE:
            print("Remove subscription to %d:\n  %s" \
                      % (self.port, PubSubPacket.bintohexid(ps_packet.rfid)))
            
#            rfid = ps_packet.rfid[:]
            rfid_i1 = PubSubPacket.RFID_OFFSET
            rfid_i2 = rfid_i1 + PubSubPacket.FID_LEN
            rfid = ps_packet.packet[rfid_i1:rfid_i2].tostring()
            if rfid in self.subscriptions:
                l = self.subscriptions[rfid]
                l[0] -= 1
                if l[0] == 0:
                    # Remove this item
                    del self.subscriptions[rfid]
                    
                    # Reconstruct FId
#                    self.ps_packet.fid[:] = self.initial_fid
#                    addr = self.ps_packet.packet_addr
#                    offset = PubSubPacket.FID_OFFSET
                    fid_i1 = self.ps_packet.FID_OFFSET
                    fid_i2 = fid_i1 + self.ps_packet.FID_LEN
                    self.ps_packet.packet[fid_i1:fid_i2] = \
                        array.array("c", self.initial_fid)
                    length = PubSubPacket.FID_LEN
                    for value in self.subscriptions.values():
#                        fidor(addr, offset, value[2], 0, length)
                        PubSubPacket.arror(self.ps_packet.packet, fid_i1,
                                           value[1], 0,
                                           length)
            print("FId:\n  %s" \
                      % PubSubPacket.bintohexid(self.ps_packet.fid))


class ClientProxy(object):
    
    def __init__(self, port, sub_fid, forwarder):
        self.address = "localhost"
        self.port = port
        self.sub_fid = sub_fid
        self.forwarder = forwarder
        
        # Create socket
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    
    def send(self, ps_packet):
        if ps_packet.packet_type == PubSubPacket.PUB_TYPE:
#            data_len = struct.unpack_from("I", ps_packet.len, 0)[0]
            data_len = struct.unpack_from("I",
                                          ps_packet.packet,
                                          PubSubPacket.LEN_OFFSET)[0]
#            payload = buffer(ps_packet.payload, 0, data_len)
            payload = buffer(ps_packet.packet,
                             PubSubPacket.PAYLOAD_OFFSET,
                             data_len) 
            
            return self.udp_socket.sendto(payload, (self.address, self.port))
    
    def subscribe(self):
        print("Subscribe")
        
        # Create packet
        initial_packet = \
            (PubSubPacket.PAYLOAD_OFFSET+PubSubPacket.MD_LEN)*"\x00"
        self.ps_packet = PubSubPacket(initial_packet)
        
        # Set type
#        struct.pack_into("B", self.ps_packet.type, 0, PubSubPacket.SUB_TYPE)
        struct.pack_into("B", self.ps_packet.packet, PubSubPacket.TYPE_OFFSET,
                         PubSubPacket.SUB_TYPE)
        # Set TTL
        self.ps_packet.packet[PubSubPacket.TTL_OFFSET] = "\xff"
        # Set FId
#        self.ps_packet.fid[:] = self.sub_fid
        fid_i1 = PubSubPacket.FID_OFFSET
        fid_i2 = fid_i1 + PubSubPacket.FID_LEN
        self.ps_packet.packet[fid_i1:fid_i2] = \
            array.array("c", self.sub_fid)
        # Set RId (port)
#        struct.pack_into("I", self.ps_packet.rid, 0, self.port) # XXX
        struct.pack_into("I", self.ps_packet.packet, PubSubPacket.RID_OFFSET,
                         self.port) # XXX
        
        self.forwarder.handle_out(self.ps_packet)
    
    def unsubscribe(self):
#        struct.pack_into("B", self.ps_packet.type, 0, PubSubPacket.UNSUB_TYPE)
        struct.pack_into("B", self.ps_packet.packet, PubSubPacket.TYPE_OFFSET,
                         PubSubPacket.UNSUB_TYPE)
        self.forwarder.handle_out(self.ps_packet)
    

class EthInterface(object):
    BUFSIZE = 1514
    BC_ADDR = 6*"\xff"
    
    def __init__(self, interface, lid, proto, forwarder):
        self.interface = interface
        self.lid = lid; self.lid_addr = self.lid.buffer_info()[0]
        self.proto = proto
        self.forwarder = forwarder
        
        self.eth_socket = \
            socket.socket(socket.AF_PACKET, socket.SOCK_RAW, proto)
        self.eth_socket.bind((interface, proto))
        
        if_name, if_proto, pkt_type, hw_type, hw_addr = \
            self.eth_socket.getsockname()
        self.hw_addr = hw_addr
        
        self.recvbuf = array.array("c", self.BUFSIZE*"\x00")
        self.ps_packet = PubSubPacket(self.recvbuf)
#        self.packet_addr = self.ps_packet.packet_addr
    
    def fileno(self):
        return self.eth_socket.fileno()
        
    def send(self, ps_packet):
        # Set Ethernet II header
        struct.pack_into("!6s6sH", ps_packet.packet, 0,
                         self.BC_ADDR, self.hw_addr, self.proto)
        
#        print("Sending out %d bytes on interface %s" \
#                  % (len(ps_packet.packet), self.interface))
        packet = buffer(ps_packet.packet, 0, ps_packet.packet_len)
        return self.eth_socket.sendall(packet)
    
    def _read(self):
        return self.eth_socket.recv_into(self.recvbuf, self.BUFSIZE)
    
    def receive(self):
        self.ps_packet.packet_len = self._read()
        
        self.ps_packet.packet_type = \
            struct.unpack_from("B",
                               self.ps_packet.packet,
                               PubSubPacket.TYPE_OFFSET)[0]
#            struct.unpack_from("B", self.ps_packet.type, 0)[0]
        
        if self.ps_packet.packet_type in \
                (PubSubPacket.SUB_TYPE, PubSubPacket.UNSUB_TYPE):
            # Add LId to packet
#            fidor(self.packet_addr, PubSubPacket.RFID_OFFSET,
#                  self.lid_addr, 0,
#                  PubSubPacket.FID_LEN)
            PubSubPacket.arror(self.ps_packet.packet, PubSubPacket.RFID_OFFSET,
                               self.lid, 0,
                               PubSubPacket.FID_LEN)
        
        return self.forwarder.handle_in(self.ps_packet)
    

class Forwarder(object):
    def __init__(self):
        self.in_handlers = {} # {port/RId: [handler, ...], ...}
        self.out_handlers = {}  # {LId: handler, ...}
    
    def add_in_handler(self, port, handler):
        if port not in self.in_handlers:
            self.in_handlers[port] = set([])
        self.in_handlers[port].add(handler)
    
    def set_out_handler(self, lid, handler):
        self.out_handlers[lid] = handler
    
    def handle_in(self, ps_packet):
#        port = struct.unpack_from("I", ps_packet.rid, 0)[0]
        port = struct.unpack_from("I",
                                  ps_packet.packet,
                                  PubSubPacket.RID_OFFSET)[0]
        if port in self.in_handlers:
            for handler in self.in_handlers[port]:
                handler(ps_packet)
    
    def handle_out(self, ps_packet):
        for lid, handler in self.out_handlers.items():
#            if fidand(ps_packet.packet.buffer_info()[0],
#                      PubSubPacket.FID_OFFSET,
#                      lid,
#                      0,
#                      PubSubPacket.FID_LEN):
            if Forwarder.fidand(ps_packet.packet, PubSubPacket.FID_OFFSET,
                                lid, 0,
                                PubSubPacket.FID_LEN):
                handler(ps_packet)
    
    @staticmethod
    def fidand(fid, fid_offset, lid, lid_offset, length):
        i = fid_offset; j = lid_offset
        for k in xrange(0, length):
            fid_i = ord(fid[i]); lid_j = ord(lid[j])
            if fid_i & lid_j != lid_j:
                return False
            i += 1; j += 1
        return True


def _main():
    import getopt
        
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "s:c:f:g:i:e:h", ["help"])
        #print(opts, args)
    except getopt.GetoptError, err:
        print(str(err))
        sys.exit(1)
    
    server_proxies = set([])
    client_proxies = set([])
    default_pub_fid = PubSubPacket.FID_LEN*"\x00"
    default_sub_fid = PubSubPacket.FID_LEN*"\x00"
    interfaces = set([])
    proto = 0xacdc
    forwarder = Forwarder()
    
    # Scan for help
    for o, a in opts:
        if o == "-h" or o == "--help":
            name = str(__file__).split("/")[-1]
            print("%s %s" % (name,
                             "\n    -s server_port1,pub_fid1 ..."
                             "\n    -c client_port1,sub_fid1 ..."
                             "\n    -f default_pub_fid"
                             "\n    -g default_sub_fid"
                             "\n    -i interface1,lid1 ..."
                             "\n    -e ethertype"
                             "\n    -h (help)\n"
                             "\n  Example 1: " \
                                 "%s -i eth0,::01 -i eth1,::02 -s 1234 -s 1235"
                             "\n  Example 2: " \
                                 "%s -i eth0,::04 -c 1234,::a4 -c 6543,::b4"
                             % (name, name)))
            sys.exit()
    
    # Scan for default FIds
    for o, a in opts:
        if o == "-f":
            default_pub_fid = PubSubPacket.hextobinid(a)
        if o == "-g":
            default_sub_fid = PubSubPacket.hextobinid(a)
    
    # Scan for Ethertype
    for o, a in opts:
        if o == "-e":
            proto = int(a, 16)
    
    # Scan for ports and interfaces
    for o, a in opts:
        if o in ("-f", "-g", "-e"):
            pass
        elif o == "-s":
            al = a.split(",")
            port = int(al[0])
            pub_fid = (PubSubPacket.hextobinid(al[1]) if (len(al) >= 2)
                       else default_pub_fid)
            server_proxy = ServerProxy(port, pub_fid, forwarder)
            server_proxies.add(server_proxy)
            # Server receives subscription updates
            forwarder.add_in_handler(port, server_proxy.update_subscription)
        elif o == "-c":
            al = a.split(",")
            port = int(al[0])
            sub_fid = (PubSubPacket.hextobinid(al[1]) if (len(al) >= 2)
                       else default_sub_fid)
            client_proxy = ClientProxy(port, sub_fid, forwarder)
            client_proxies.add(client_proxy)
            # Client receives data
            forwarder.add_in_handler(port, client_proxy.send)
        elif o == "-i":
            al = a.split(",")
            interface = al[0]
            lid = array.array("c", PubSubPacket.hextobinid(al[1]))
            eth_interface = EthInterface(interface, lid, proto, forwarder) 
            interfaces.add(eth_interface)
            forwarder.set_out_handler(lid.tostring(), eth_interface.send)
        else:
            assert False, "Unhandled option: %r" % o
    
    for client_proxy in client_proxies:
        client_proxy.subscribe()        
    
    def signal_handler(signal_number, stack_frame):
        raise KeyboardInterrupt()
    signal.signal(signal.SIGINT, signal_handler)
    
    recv_set = server_proxies.union(interfaces)
    print("Receiving packets on the following ports and interfaces:")
    for receiver in recv_set:
        if isinstance(receiver, ServerProxy):
            print("  udp: %d" % receiver.port)
        elif isinstance(receiver, EthInterface):
            print("  eth: %s" % receiver.interface)
    if len(recv_set) == 0:
        print("Warning: no input ports or interfaces specified "
              "(see -h for help)")
    try:
        while True:
            l = select.select(recv_set, (), ())
            for receiver in l[0]:
                receiver.receive()
    except (SystemExit, KeyboardInterrupt):
        for client_proxy in client_proxies:
            client_proxy.unsubscribe()
        time.sleep(0.5)
    
if __name__ == "__main__":
    _main()
