/*
* Copyright (c) 2009, Nomadiclab, Oy LM Ericsson Ab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of BSD
* license.
* 
* See LICENSE and COPYING for more details.
*/

module output_port_selector
  #(parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH=DATA_WIDTH/8,
    parameter NUM_OUTPUT_QUEUES = 8,
    parameter UDP_REG_SRC_WIDTH = 2,
    parameter FILTER_SIZE 		= 64,
    parameter FILTER_DEPTH		= 4, // can be modified
    parameter LINKS			= 4, // can be modified
    parameter MAX_IDS			= 4, // can be modified
    parameter MAX_LITS			= 4) // can be modified
    (
	
     // Connection to output queues
     output reg [NUM_OUTPUT_QUEUES-1:0]  	out_selection_data,
     output reg  				out_selection_ready,
     output reg [7:0]				out_ttl,
     output reg [DATA_WIDTH-1:0]		combined_zFilter,
     input [log2(DATA_WIDTH)-1:0]		collected_index,
     
     // Incoming data from output_queues in
     input      [DATA_WIDTH-1:0]      		in_selection_data,
     input      [CTRL_WIDTH-1:0]      		in_selection_ctrl,
     input                         		in_wr,
     
     // Connection to registers
     // --- interface to udp_reg_grp
     input                               	reg_req_in,
     input                               	reg_ack_in,
     input                               	reg_rd_wr_L_in,
     input  [`UDP_REG_ADDR_WIDTH-1:0]    	reg_addr_in,
     input  [`CPCI_NF2_DATA_WIDTH-1:0]   	reg_data_in,
     input  [UDP_REG_SRC_WIDTH-1:0]      	reg_src_in,
     
     output reg                              	reg_req_out,
     output reg                              	reg_ack_out,
     output reg                              	reg_rd_wr_L_out,
     output reg [`UDP_REG_ADDR_WIDTH-1:0]    	reg_addr_out,
     output reg [`CPCI_NF2_DATA_WIDTH-1:0]   	reg_data_out,
     output reg [UDP_REG_SRC_WIDTH-1:0]      	reg_src_out,
     
     input 					clk,
     input                            		reset
     );
   
   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2
   
   // Amount of required parallel filters
   localparam FILTER_NUMBER         = LINKS*(MAX_IDS);
   
   localparam DEPTH_WIDTH = log2(FILTER_DEPTH);
   localparam INDEX_WIDTH = log2(MAX_LITS);   
   localparam INDEX_LENGTH = log2(FILTER_DEPTH)+log2(MAX_LITS);
   
   // -------- state parameters for the header parser
   // wait for new header
   // wait for data and make port selection
   // wait for end of packets
   localparam NUM_INPUT_STATES 		= 3;
   localparam IN_WAIT_HEADER 		= 1;
   localparam IN_WAIT_PKT_DATA   	= 2;
   localparam IN_WAIT_EOP         	= 4;
   
   localparam DEFAULT_MAX_ONES		= 40;
      
	reg collect_zFilter;
	reg [DATA_WIDTH-1:0] 		collected_zFilter [(FILTER_DEPTH)-1:0];

	//set desired pieces of collected zFilter to output for output_queues 
	always @(posedge clk) begin
	    if(collect_zFilter) begin
		    if(collected_index < 1) begin
		        combined_zFilter[47:0] = collected_zFilter[0][63:16];
		        combined_zFilter[63:48] = 16'h0;
		    end
	    	else if (collected_index >= 4) begin
		        combined_zFilter[47:0] = 48'h0;
		        combined_zFilter[63:48] = collected_zFilter[3][47:0];
		    end
			else begin
		    	combined_zFilter[47:0] = collected_zFilter[collected_index][63:16];
		        combined_zFilter[63:48] = collected_zFilter[collected_index-1][15:0];
		    end
		end
		else
    		combined_zFilter = 64'h0;
	end
      
   // state registers
   reg [NUM_INPUT_STATES-1:0] input_state;
   reg [NUM_INPUT_STATES-1:0] input_state_next;
   
   // Storage for making port selection
   reg [MAX_IDS-1:0] 	      selections[LINKS-1:0];
   wire [NUM_OUTPUT_QUEUES-1:0] selection;
   
   // Indicator if selection is finished and output should be set
   reg 				out_selection_ready_next;
   
   // Internal request to start and stop packet counter
   reg 				count_header;
   reg [2:0] 			header_counter;
   
   // Debug register
   reg [DATA_WIDTH-1:0] 	debug_reg;
   
   // Current filters, values are updatet from id_store, once
   // per clock cyckle when counting is on
   wire [FILTER_SIZE-1:0] 	filt[FILTER_NUMBER-1:0];
   
   // Link table index
   reg [2:0] 			lit;
   
   // Source port
   reg [7:0] 			incoming_port;	
   wire [7:0] 			src;
   
   // Ethertype
   reg [15:0] 			ethertype;
   
   // Bitcounter	
   wire [7:0] 			count;
   reg [7:0] 			last_count;
   reg [7:0] 			max_ones;
   
   // TTL
   reg [7:0] 			ttl;
   
   // Counter that counts the amount of ones
   bit_count counter (
		      .in(in_selection_data),
		      .out(count)
		      );
   
   //Connectors for id_storages
   wire 			read_en;
   wire [INDEX_LENGTH-1:0] 	read_index;
   wire [FILTER_NUMBER-1:0] 	store_write;
   wire [FILTER_NUMBER-1:0] 	store_read;
   wire [DATA_WIDTH/2-1:0] 	store_write_data;
   wire [DATA_WIDTH/2-1:0] 	store_read_data[FILTER_NUMBER-1:0];
   wire [INDEX_LENGTH-1:0] 	reg_index;
   wire 			index_reg_hig;
   wire [FILTER_NUMBER-1:0] 	data_reg_rdy;
   
   assign 			read_en = count_header & header_counter >= 2;
   
   // use table index straight from the packet if in the header of the packet, othervise use stored one
   assign 			read_index = (header_counter == 2) ? 
				(header_counter-2 + FILTER_DEPTH*in_selection_data[55:48]) : (header_counter-2 + lit*FILTER_DEPTH);
   
   // Generate enought id_storages
   generate
      genvar 			gi;
      for(gi=0; gi<FILTER_NUMBER; gi=gi+1) begin: filter_storage
   	 id_store
   	   #(.DATA_WIDTH(DATA_WIDTH),
	     .DEPTH(FILTER_DEPTH*MAX_LITS),
	     .INDEX_LENGTH(INDEX_LENGTH)) 
	     id_store (
   		       .read(read_en),
		       .index(read_index),
		       .data(filt[gi]),
		       .write_reg(store_write[gi]),
		       .read_reg(store_read[gi]),
		       .index_reg(reg_index),
		       .index_reg_hig(index_reg_hig),
		       .data_reg_in(store_write_data),
		       .data_reg_out(store_read_data[gi]),
		       .data_reg_rdy(data_reg_rdy[gi]),
 		       .clk(clk)
   		       );
      end
   endgenerate

   // Header parse statemachine
   always @(*) begin
      
      out_selection_ready_next = out_selection_ready;
      input_state_next = input_state;
      
      case(input_state)
     	IN_WAIT_HEADER: begin
     	   // Wait for the begining of a header 
	   count_header = 0;
  	   if(in_wr && in_selection_ctrl == `IO_QUEUE_STAGE_NUM) begin
  	      input_state_next = IN_WAIT_PKT_DATA;
  	      out_selection_ready_next = 0;
  	   end
  	end
  	
  	IN_WAIT_PKT_DATA: begin
  	   // Pass NetFPGA headers and wait for data start, in_selection_ctrl == 0
	   count_header = 0;
  	   if(in_wr && in_selection_ctrl == 0) begin
  	      if(header_counter >= FILTER_DEPTH + 3) begin
		 input_state_next = IN_WAIT_EOP;
		 out_selection_ready_next = 1;
  	      end
  	      else
  	     	count_header = 1;
  	   end
  	end
 	
 	IN_WAIT_EOP: begin
	   count_header = 0;
	   // Wait for the end of a packet, in_ctrl != 0
  	   if(in_wr && in_selection_ctrl != 0) begin
  	      input_state_next = IN_WAIT_HEADER;
  	      out_selection_ready_next = 0;
  	   end
 	end
 	
 	default: begin // We should not get in here
 	   count_header = 0;
 	   input_state_next = IN_WAIT_HEADER; 
 	end
      endcase
   end

   assign src = in_selection_data[`IOQ_SRC_PORT_POS+15:`IOQ_SRC_PORT_POS];
   
   // State changes for header state
   always @(posedge clk) begin
      if(reset) begin
      	 out_selection_data 	<= 8'b00000000;
      	 out_selection_ready 	<= 0;
      	 input_state 		<= IN_WAIT_HEADER;
      end
      else begin
         // Set output if selection just finished
         if(out_selection_ready_next && !out_selection_ready) begin
            if(ttl > 0 && last_count <= max_ones
	       && ethertype == 16'hacdc) begin //XXX define ethertype
               out_selection_data 	<= selection;
	       out_ttl 			<= ttl - 1;
	    end
	    else
	      out_selection_data 	<=8'b00000000;
	 end
	 
	 // Get correct source port
	 if(in_wr && in_selection_ctrl == `IO_QUEUE_STAGE_NUM) begin
	    incoming_port 	<= src;
	 end
	 
         out_selection_ready <= out_selection_ready_next;
         input_state 		<= input_state_next;
      end      
   end
   
   // Combine forwarding result from all filter
   // and map those to actual ports
   generate
      genvar gj;
      for(gj=0;gj<4;gj=gj+1) begin: select
	 assign selection[gj*2] = |selections[gj];
      end
   endgenerate
   //assign selection[1] = |selections[4];
   // //packets to cpu only through one if
   assign 	selection[1] = 0;
   assign 	selection[3] = 0;
   assign 	selection[5] = 0;
   assign 	selection[7] = 0;
   
   // --- Make actual zFiltering
   integer 	i,j;
   always @(posedge clk) begin
      if(reset) begin
	 header_counter <= 0;
	 for(i=0;i<LINKS;i=i+1) begin
	    selections[i] <= ~0;
	 end
      end
      else begin
	 if(count_header) begin //if incoming filter data
	    if(in_wr) begin //if incoming data has something
	       if(header_counter == 1) begin
		  // Reset selections	to one. Later all mismaches
		  // are set to zero 
		  for(i=0;i<LINKS;i=i+1) begin
		     selections[i] <= ~0;
		  end
		  // Set incoming port to zero and prevent
		  // sending of the packet back to sender
		  selections[incoming_port/2] <= 0;
   		  ethertype <= in_selection_data[31:16];
		  last_count <= 0;
		  debug_reg[31:0] <= in_selection_data[31:0];
		  debug_reg[63:32] <= in_selection_data[15:8];
		  if(in_selection_data[7:0] == 8'h05 || in_selection_data[7:0] == 8'h03 ||
	       		in_selection_data[7:0] == 8'h02)
	       			collect_zFilter <= 1;
	       		else
	       		    collect_zFilter <= 0;
	       end  
	       else if(header_counter == 2) begin
		  // Get ttl and filter table index from packet
		  ttl <= in_selection_data[63:56];
		  lit <= in_selection_data[55:48];
	       end
	       else if (header_counter>2) begin
		
		  // Make forwarding decision
		  for(i=0;i<LINKS;i=i+1) begin
		     for(j=0;j<MAX_IDS;j=j+1) begin
			selections[i][j] <= (selections[i][j]) & ((in_selection_data &
					      filt[(MAX_IDS)*i+j]) == filt[(MAX_IDS)*i+j]);
		     end
		  end
		  
		  collected_zFilter[(header_counter-3)] <= filt[(incoming_port/2)*(MAX_IDS)];
		  last_count <= count+last_count;
	       end
	       
  	       header_counter <= header_counter+1;
	    end //if(in_wr)
	 end //if(count_header)
	 else begin
	    header_counter <= 0;
	 end	
      end
   end
   
   
   // --- Register access 
   // ADDR_WIDTH is 10
   localparam ADDR_WIDTH = `UDP_BLOCK_SIZE_1k_REG_ADDR_WIDTH;
   
   wire [ADDR_WIDTH-1:0] addr;
   wire [`UDP_REG_ADDR_WIDTH-`ZF_OQ_REG_ADDR_WIDTH - 1:0] tag_addr;
   
   wire 						  tag_hit;
   
   wire 						  status_tag_hit;
   wire [`ZF_OQ_REG_ADDR_WIDTH-1:0] 			  status_reg_addr;
   
   // -------------- Logic --------------------
   assign 						  addr = reg_addr_in[`ZF_OQ_REG_ADDR_WIDTH-1:0];
   assign 						  tag_addr = reg_addr_in[`UDP_REG_ADDR_WIDTH - 1:`ZF_OQ_REG_ADDR_WIDTH];
   
   assign 						  tag_hit = tag_addr == `ZF_OQ_BLOCK_TAG;
   
   assign 						  status_tag_hit = tag_addr == `ZF_STATUS_OQ_BLOCK_TAG;
   assign 						  status_reg_addr = reg_addr_in[`ZF_OQ_REG_ADDR_WIDTH-1:0];
   
   wire [ADDR_WIDTH-DEPTH_WIDTH-1:0] 			  filter_num;
   wire [INDEX_WIDTH-1:0] 				  table_index;
   wire [DEPTH_WIDTH-1:0] 				  dep_index;
   
   //port and link index
   assign 						  filter_num 		= addr[ADDR_WIDTH-1 : DEPTH_WIDTH+1+INDEX_WIDTH];
   //id number
   assign 						  table_index 		= addr[DEPTH_WIDTH+INDEX_WIDTH : DEPTH_WIDTH+1];
   //dep index
   assign 						  dep_index		= addr[DEPTH_WIDTH : 1];
  
   
   // Temporary registers to store write data requests
   reg [`CPCI_NF2_DATA_WIDTH-1:0] 			  write_data;
   reg [FILTER_NUMBER-1:0] 				  write_id;
   reg [DEPTH_WIDTH+INDEX_WIDTH-1:0] 			  write_dep;
   reg 							  write_high;

   // Temporary registers to store read data requests	
   reg [FILTER_NUMBER-1:0] 				  read_id;
   reg [ADDR_WIDTH-DEPTH_WIDTH-1:0] 			  read_num;
   reg [DEPTH_WIDTH+INDEX_WIDTH-1:0] 			  read_dep;
   reg 							  read_high;
   reg [`UDP_REG_ADDR_WIDTH-1:0] 			  read_addr_in;
   reg [UDP_REG_SRC_WIDTH-1:0] 				  read_src_in;
   
   // Connect reqister handler to ID stores
   assign 						  store_write = write_id;
   assign 						  store_read  = read_id;
   assign 						  store_write_data = write_data;
   
   assign 						  index_reg_hig 	= |write_id ? write_high : read_high;
   assign 						  reg_index 		= |write_id ? write_dep : read_dep;
   
   // Register I/O
   always @(posedge clk) begin
      reg_rd_wr_L_out <= reg_rd_wr_L_in;
      reg_addr_out <= reg_addr_in;
      reg_src_out <= reg_src_in;
      write_id <= 0;
      //read_id <= 0;
     
      if( reset ) begin
         reg_req_out <= 1'b0;
         reg_ack_out <= 1'b0;
         reg_data_out <= 'h0;
         max_ones 	<= DEFAULT_MAX_ONES;
      end
      else begin
         if (reg_req_in && tag_hit) begin
            reg_data_out <= 32'hcafe_babe;
	    
            if (!reg_rd_wr_L_in) begin  
               //store write reguest data and indexes and make request
               write_data 				<= reg_data_in;
               write_id 				<= 0;
               write_id[filter_num]		<= 1;
               write_high 				<= ~addr[0];
               write_dep 				<= dep_index + table_index*FILTER_DEPTH;
               
               //set correct outputs
               reg_data_out 	<= reg_data_in;
               reg_ack_out 	<= 1'b1;
               reg_req_out 	<= 1'b1;
            end
            else begin
               //store read reguest data and indexes and make request
	       read_id 				<= 0;
	       read_id[filter_num]		<= 1;
	       read_addr_in 			<= reg_addr_in;
	       read_src_in 			<= reg_src_in;
	       read_high			<= ~addr[0];
               read_num 			<= filter_num;
 	       read_dep 			<= dep_index + table_index*FILTER_DEPTH;
	  
	       //set outputs
	       reg_ack_out 	<= 1'b0;
	       reg_req_out 	<= 1'b0;
	       reg_data_out 	<= 'b0;
	       reg_addr_out 	<= 'b0;
            end
	    // read request takes now 3 cyckles
	    // and write 2 cyckles
   	    
         end
         else if (reg_req_in && status_tag_hit) begin
            // Handle configuration and debug register access
            if(status_reg_addr==0) begin //return configuration
	       reg_data_out[7:0] 	<= LINKS;
	       reg_data_out[15:8] 	<= MAX_IDS;
	       reg_data_out[23:16]	<= MAX_LITS;
	       reg_data_out[31:24]	<= FILTER_DEPTH;
	    end
	    else if(status_reg_addr==1) begin //max ones
	       if(!reg_rd_wr_L_in)
	         max_ones <= reg_data_in;
	       else
	         reg_data_out <= max_ones;
	    end
	    else if(status_reg_addr==2) begin //info from last packet
	       reg_data_out[7:0] 	<= last_count;
	       reg_data_out[15:8] 	<= lit;
	       reg_data_out[23:16]	<= ttl;
	       reg_data_out[31:24]	<= incoming_port/2;
	       
	    end
	    else if(status_reg_addr==3) begin // zFilter results from last selection
	       for(i=0;(i<LINKS) && (i*MAX_IDS < `CPCI_NF2_DATA_WIDTH-MAX_IDS);i=i+1) begin
	          reg_data_out[((i)*MAX_IDS)+:MAX_IDS] <= selections[i];
	       end
	    end
	    else if(status_reg_addr==4) begin //debug register
 	       reg_data_out 	<= debug_reg[31:0];
 	       //reg_data_out 	<= combined_zFilter[31:0];
	    end else
	      reg_data_out 	<= debug_reg[63:32];
	      //reg_data_out 	<= combined_zFilter[63:32];
	 
            reg_ack_out 	<= 1'b1;
      	    reg_req_out 	<= 1'b1;
      	    reg_addr_out 	<= reg_addr_in;
      	    reg_src_out 	<= reg_src_in;
      	    reg_rd_wr_L_out <= 1;    	       	    
         end
     	 else begin
     	    //nothing for us so just forward everything
            reg_ack_out 	<= reg_ack_in;
            reg_data_out 	<= reg_data_in;
	    reg_req_out 	<= reg_req_in;
         end
         
         if(data_reg_rdy[read_num]) begin 
	    reg_data_out 	<= store_read_data[read_num];
            reg_ack_out 	<= 1'b1;
      	    reg_req_out 	<= 1'b1;  
      	    reg_addr_out 	<= read_addr_in; 
      	    reg_src_out 	<= read_src_in;  
      	    reg_rd_wr_L_out <= 1;
      	    read_id 		<= 0;    	
         end
         
      end // else: !if( reset )
   end // always @ (posedge clk)
   
endmodule
