#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <stdint.h>

#include "nflib/api.h"
#include "common/reg_defines.h"
#include "common/zf_defines.h"
#include "configparser.h"
#include "psm.h"
#include "helpers.h"

extern struct global_args g;

/* create new nf2_cmd structure, and allocate its *target* member memory of
 * size bytes.
 */
struct nf2_cmd *newcmd(int cmd, int size)
{
        struct nf2_cmd *ret;

        ret = calloc(1, sizeof(*ret));
        if (!ret) {
                global_error_exit("out of memory");
        }

        ret->target = calloc(1, size);
        if (!ret->target) {
                global_error_exit("out of memory");
        }
        ret->len = size;
        ret->cmd = cmd;
        return ret;
}

struct nf2_cmd *freecmd(struct nf2_cmd *chain)
{
        struct nf2_cmd *iter, *next;
        next = chain;

        do {
                iter = next;
                next = iter->next;
                free(iter->target);
                free(iter);
        } while(next);

        return NULL; /* enables:  cmd = freecmd(cmd); */
}


/* copy suitable filters (from configuration file) to their respective 
 * commands (structures).
 * We could optimize that instead of doing memcpy we would just link the
 * commands to use the configuration file memory. This could break later,
 * so for simplicity, we'll just use double memory.
 */
static void getinput(struct psm_conf *spex, struct nf2_cmd *cmd, 
                     ztable table)
{
        int port, link, tid;
        int num, n, err = 0;
        zfilter z;
        uint8_t null_zfilter[ZFILTER_SIZE] = {0};

        port = spex->port == INDEX_ANY ? 0 : spex->port;
        link = spex->link == INDEX_ANY ? 0 : spex->link;
        tid = spex->tid   == INDEX_ANY ? 0 : spex->tid;

        num = cmd->len / ZFILTER_SIZE;
        n = num;
        while(num-- > 0 && err == 0) {
                z = GETFILTER(table, tid, port, link);
                if (z == NULL)
                        memcpy(cmd->target+(n-(num+1))*ZFILTER_SIZE, 
                               null_zfilter, ZFILTER_SIZE);
                else
                        memcpy(cmd->target+(n-(num+1))*ZFILTER_SIZE, z, 
                               ZFILTER_SIZE);
                NEXTFILTER(tid, port, link, err);
        }
}

#define TRAILER do {                          \
        cmd->addr = calc_nf2_address(spex);   \
        add_to_nf2list(&prev, cmd);           \
        if (mode == NF_WRITEZF)               \
              getinput(spex, cmd, g.rttable); \
} while(0)

/* PTL = Port, Tableid, Link (shorthands for functions below)*/
#define ZF_PTL (spex->port & spex->tid & spex->link)
#define ZF_PT  (spex->port & spex->tid             )
#define ZF_PL  (spex->port             & spex->link)
#define ZF_TL  (             spex->tid & spex->link)

#define ZF_IDX_PTL (g.nfconf.num_links*g.nfconf.num_ports*g.nfconf.num_tids)
#define ZF_IDX_PL  (g.nfconf.num_links*g.nfconf.num_ports                  )

struct nf2_cmd *ZFilterCmd(struct psm_conf *spex)
{
        struct nf2_cmd *cmd = NULL, *prev = NULL;
        int i,j;
        int mode = NF_READZF;

        if (GET_CMD(spex) == CMD_SET) {

                if (spex->file)
                        global_error_exit("Reading from stdio is not "
                                          "supported\n");
        
                if (!spex->prebuffer)
                        global_error_exit("No input data found\n");

                g.rttable = cp_parse(spex->prebuffer);
                if (g.rttable == NULL)
                        global_error_exit("Unable to parse zfilter input\n");

                mode = NF_WRITEZF;
        }


        if ((ZF_PTL) == INDEX_ANY) {
                for(i=0;i<g.nfconf.num_tids;i++) {
                        cmd = newcmd(mode, ZFILTER_SIZE * ZF_IDX_PL);
                        spex->tid = i;
                        TRAILER;
                }
        } else if ((ZF_PT) == INDEX_ANY) {
                /* get all zfilters for all ports and tids on a specific link*/
                for(i=0; i<g.nfconf.num_ports;i++) {
                        for(j=0;j<g.nfconf.num_tids;j++) {
                                cmd = newcmd(mode, ZFILTER_SIZE);
                                spex->port = i;
                                spex->tid = j;
                                TRAILER;
                        }
                }

        } else if ((ZF_PL) == INDEX_ANY) {
                /* get all zfilters for all ports and links on a specific tid*/
                cmd = newcmd(mode, ZFILTER_SIZE * ZF_IDX_PL);
                TRAILER;

        } else if ((ZF_TL) == INDEX_ANY) {
                /* get all zfilters for all tids and links on a specific port*/
                for(i=0; i<g.nfconf.num_tids;i++) {
                        cmd = newcmd(mode, ZFILTER_SIZE * 
                                     g.nfconf.num_links);
                        spex->tid = i;
                        TRAILER;
                }

        } else if (spex->link == INDEX_ANY ||
                   spex->tid  == INDEX_ANY ||
                   spex->port == INDEX_ANY) 
        {
                int upidx = g.nfconf.num_ports;
                int lk = 0, pt = 1;

                if (spex->link == INDEX_ANY) {
                        lk = 1; pt = 0; upidx = g.nfconf.num_links;
                } else if (spex->tid == INDEX_ANY) {
                        lk = 0; pt = 0; upidx = g.nfconf.num_tids;
                } 
                /* else { lk = 0; pt = 1; updidx = g.nfconf.num_ports; } */

                        
                for(i=0; i<upidx; i++) {
                        cmd=newcmd(mode, ZFILTER_SIZE);
                        if (lk)
                                spex->link = i;
                        else if (pt)
                                spex->port = i;
                        else
                                spex->tid = i;

                        TRAILER;
                }

        } else {
                /* specific port, link and tid request */
                cmd = newcmd(mode, ZFILTER_SIZE);
                TRAILER;

        }
        return (cmd->first ? cmd->first : cmd);
}

#undef TRAILER

/* Issue commands to read statistics */
struct nf2_cmd *ZFilterStatsCmd(struct psm_conf *spex)
{
        int stats[] = {
                TX_QUEUE_0_NUM_PKTS_SENT_REG,
                TX_QUEUE_0_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_0_NUM_PKTS_STORED_REG,
                RX_QUEUE_0_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_0_NUM_PKTS_DROPPED_FULL_REG,
                RX_QUEUE_0_NUM_PKTS_DROPPED_BAD_REG,

                TX_QUEUE_1_NUM_PKTS_SENT_REG,
                TX_QUEUE_1_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_1_NUM_PKTS_STORED_REG,
                RX_QUEUE_1_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_1_NUM_PKTS_DROPPED_FULL_REG,
                RX_QUEUE_1_NUM_PKTS_DROPPED_BAD_REG,

                TX_QUEUE_2_NUM_PKTS_SENT_REG,
                TX_QUEUE_2_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_2_NUM_PKTS_STORED_REG,
                RX_QUEUE_2_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_2_NUM_PKTS_DROPPED_FULL_REG,
                RX_QUEUE_2_NUM_PKTS_DROPPED_BAD_REG,

                TX_QUEUE_3_NUM_PKTS_SENT_REG,
                TX_QUEUE_3_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_3_NUM_PKTS_STORED_REG,
                RX_QUEUE_3_NUM_BYTES_PUSHED_REG,
                RX_QUEUE_3_NUM_PKTS_DROPPED_FULL_REG,
                RX_QUEUE_3_NUM_PKTS_DROPPED_BAD_REG,
        };

        struct nf2_cmd *cmd, *prev = NULL;
        int i,j;
        
        /* There are 6 statistics per port */

        if (spex->port == INDEX_ANY) {
                for(i=0;i<g.nfconf.num_ports;i++) {
                        for(j=0;j<6;j++) {
                                cmd = newcmd(NF_READ, 4);
                                cmd->addr = (int)stats[i*6+j];
                                cmd->tag = i*6+j;
                                add_to_nf2list(&prev, cmd);
                        }
                }
        } else {
                for(j=0;j<6;j++) {
                        cmd = newcmd(NF_READ, 4);
                        cmd->addr = (int)stats[spex->port*6+j];
                        cmd->tag = spex->port*6+j;
                        add_to_nf2list(&prev, cmd);
                }
        }

        return (cmd->first ? cmd->first : cmd);
}

/* Write zfilters (from chain) to conf->file (stdout/real file) */
void write_zfilters(struct psm_conf *conf, struct nf2_cmd *chain)
{
        struct nf2_cmd *iter;
        int num, n, err;
        int port, tid, link;
        uint32_t daddr;
        char tmpstr[ZFILTER_SIZE*2+1];
        uint8_t *tmpzf;

        iter = chain;
        do {
                daddr = decode_nf2_address(iter->addr);
                tid   = (daddr & 0xFF000000) >> 24;
                port  = (daddr & 0x00FF0000) >> 16;
                link  = (daddr & 0x0000FF00) >> 8;

                num = iter->len / ZFILTER_SIZE;
                tmpzf = iter->target;
                while(num-- > 0) {
                        gen_string(tmpstr, tmpzf);
                        n = fprintf(conf->file, "%d:%d:%d:%s\n", tid, 
                                    port, link, tmpstr);
                        if (n < 70) {
                                dprint(LOG_ERR, "Short write! "
                                       "(%d bytes)\n", n);
                        }
                        NEXTFILTER(tid, port, link, err);
                        tmpzf+=ZFILTER_SIZE;
                }
        } while(iter = iter->next, iter != NULL && err == 0);
}

/* Write statistics from chain to file (or buffer).
 * Buffered write is used for example in "send packet" command, where
 * the statistics are used internally (and not displayed to the user) 
 */
void write_zfilter_stats(struct psm_conf *conf, struct nf2_cmd *chain)
{
        struct nf2_cmd *iter;
        uint32_t *vals;
        int port, val;
        char *strs[] = { "Packets sent:    ", 
                         "Bytes pushed:    ",
                         "Packets stored:  ",
                         "Bytes pushed:    ",
                         "Packets dropped: ",
                         "Packets bad:     " };

        iter = chain;
        if (conf->prebuffer) {
                vals = (uint32_t *)conf->prebuffer;
                do {
                        *(vals + iter->tag) = *(uint32_t *)iter->target;
                } while(iter = iter->next, iter != NULL);
        } else if (conf->file) {
          /* 6 is the amount of statistics values we are fetching per port */
                do {
                        port = iter->tag / 6;
                        val = iter->tag % 6;
                        fprintf(conf->file, "Port %d: %s %08d\n", port,
                                strs[val], *(uint32_t *)iter->target);
                } while(iter = iter->next, iter != NULL);
        } else {
                global_error_exit("No output defined!");
        }
}
