#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <stdint.h>
#include <math.h>

#include "nflib/api.h"
#include "common/reg_defines.h"
#include "common/zf_defines.h"
#include "configparser.h"
#include "psm.h"
#include "cmds.h"
#include "helpers.h"
#include "packet.h"

/* default configuration */
struct psm_conf psmconf = { CMD_NONE, INDEX_ANY, INDEX_ANY, INDEX_ANY, 
                            {0}, NULL, NULL, 16, NULL, 1, 1000, 100, 0xACDC};


/* Predefined global variables */
struct global_args g;

void usage() 
{
        printf("Usage: psm <cmd>\n");
        printf("<cmd>   = <fcmd> filter <fspex>\n");
        printf("        | send packet <pspex>\n");
        printf("<fcmd>  = get\n");
        printf("        | set\n");
        printf("<fspex> = tid       <num> <fspex>\n");
        printf("        | port      <num> <fspex>\n");
        printf("        | link      <num> <fspex>\n");
        printf("        | file <filename> <fspex>\n");
        printf("        | <empty>\n");
        printf("<pspex> = ttl        <num> <pspex>\n");
        printf("        | if      <string> <pspex>\n");
        printf("        | cnt        <num> <pspex>\n");
        printf("        | delay      <num> <pspex>\n");
        printf("        | len        <num> <pspex>\n");
        printf("        | proto      <num> <pspex>\n");
        printf("        | output <zfilter> <pspex>\n");
        printf("        | file  <filename> <pspex>\n");
        printf("        | tid        <num> <pspex>\n");
        printf("        | <empty>\n");
        exit(0);
}


int execute_one(const struct nf2_cmd *cmd)
{
        int err = -1;
        int addr_iter = cmd->addr;
        uint32_t *target_iter = (uint32_t*)cmd->target;
        int len = cmd->len;
        int cnt = 0;

        if (cmd->cmd == NF_READZF || cmd->cmd == NF_WRITEZF) {
                /* 32 bytes are required to read one zfilter */
                if (cmd->len < 32) {
                        dprint(LOG_ERR, "Zfilter error: too short (%d)\n", 
                               cmd->len);
                        return err;
                }
                err = 0;
                
                /* the double while-loops are used to read data, that is
                 * stored continuously in local memory, but is accessed
                 * from sparse hw addresses. */
                while(len > 0 && err == 0) {
                        if (cmd->cmd == NF_READZF)
                                if (SIMULATE) {
                                        dprint(LOG_DBG, "Read: [%08x] => %08x"
                                               "\n", addr_iter, target_iter);
                                        *target_iter = ntohl(0xDEADBEEF);
                                } else {
                                        err = nflib_readreg(g.nf2, 
                                                            addr_iter, 
                                                             target_iter);
                                       *target_iter = ntohl(*target_iter);
                                }
                        else 
                                if (SIMULATE) {
                                        dprint(LOG_DBG, "Write: [%08x] <= %08x"
                                               "\n", addr_iter, 
                                               htonl(*target_iter));
                                } else {
                                        err = nflib_writereg(g.nf2, 
                                                             addr_iter,
                                                             htonl(*target_iter));
                                }
                        addr_iter+=4;
                        target_iter++;
                        len -= 4; 
                        if (++cnt >= 8) {
                        /* The next line skips to the next filter.
                         * Unfortunately, due to the way HW was defined,
                         * we cannot just read next address.
                         * 32 = cnt * 4 (bytes/read) = 8 * 4
                         */
                                addr_iter+=(1 << g.nfconf.pl_off) - 32;
                                cnt = 0;
                        }
                }
        } else if (cmd->cmd == NF_READ) {
                if (cmd->len > 4) {
                        dprint(LOG_ERR, "Only 32-bit reads allowed\n");
                        return err;
                }

                err = nflib_readreg(g.nf2, addr_iter, target_iter);
        } else {

                dprint(LOG_ERR, "Unsupported command: cmd=0x%x\n", cmd->cmd);
        }

        return err;

}

void execute_chains(struct nf2_cmd *chain)
{
        struct nf2_cmd *iter;
        int cnt;
        uint32_t code;

        iter = chain;

        do { 
                if (execute_one(iter)) {
                        dprint(LOG_ERR, "Executing command chain failed\n");
                        
                        dprint(LOG_INF, "%d. command:", cnt);
                        switch(iter->cmd) {
                        case NF_READZF:
                                dprint(LOG_INF, "READZF");
                                break;
                        case NF_WRITEZF:
                                dprint(LOG_INF, "WRITEZF");
                                break;
                        case NF_READ:
                                dprint(LOG_INF, "READ");
                                break;
                        case NF_WRITE:
                                dprint(LOG_INF, "WRITE");
                                break;
                        default:
                                dprint(LOG_INF, "UNKNOWN");
                        };

                        dprint(LOG_INF, "addr=0x%x\n", iter->addr); 
                        if (IS_IN_ZF_ZONE(iter->addr)) {
                                code = decode_nf2_address(iter->addr);
                                dprint(LOG_INF, "Tid: %d, Port: %d, Link: %d\n",
                                       (code & 0xFF000000) >> 24,
                                       (code & 0x00FF0000) >> 16,
                                       (code & 0x0000FF00) >> 8);
                        }
                        /* stopping execution */
                        return;
                }
        } while(cnt++, iter = iter->next, iter != NULL);
}


void execute(struct psm_conf *conf)
{
        struct nf2_cmd *chain = NULL;

        if (IS_TGT_SET(conf, TGT_FILTER)) {
                if (GET_CMD(conf) == CMD_SET ||
                    GET_CMD(conf) == CMD_GET) {
                        chain = ZFilterCmd(conf);
                } else {
                        dprint(LOG_ERR, "Illegal command: 0x%08x\n", 
                               conf->command);
                        global_error_exit("");
                }

                g.cleanup.cmdchain = chain;
                if (!chain || chain->first != chain) {
                        global_error_exit("Consistency check for command chain "
                                          "failed");
                }

                execute_chains(chain);

                if (GET_CMD(conf) == CMD_GET && 
                    IS_TGT_SET(conf, TGT_FILTER)) {
                        /* write output to file/stdout */
                        write_zfilters(conf, chain);
                }

                freecmd(chain);
                g.cleanup.cmdchain = NULL;

        } else if (IS_TGT_SET(conf, TGT_STATS)) {
                if (GET_CMD(conf) == CMD_GET) {
                        chain = ZFilterStatsCmd(conf);
                } else {
                        global_error_exit("STATS target is only supported for"
                                          " GET command.\n");
                }

                g.cleanup.cmdchain = chain;
                execute_chains(chain);

                write_zfilter_stats(conf, chain);
                freecmd(chain);
                g.cleanup.cmdchain = NULL;

        } else if (GET_CMD(conf) == CMD_SEND) {
                /* send packet(s) */
                do_packet_send(conf);
        }


        /* done */
        return;
}

/* read configuration from hardware and setup internal variables */
void basic_setup(void)
{
        unsigned int config;
        int tid_len;
        int pl_len;
        int mask;

        if (SIMULATE)
                config = 0x04040404;
        else
                nflib_readreg(g.nf2, ZF_CONFIG_BASE, &config);

        dprint(LOG_DBG, "Read config: %08x\n", config);
        g.nfconf.num_ports = config & 0xFF;
        g.nfconf.num_links = (config & 0xFF00) >> 8;
        g.nfconf.num_tids = (config & 0xFF0000) >> 16;

        dprint(LOG_INF, "Current config: tids = %d, ports = %d, links = %d\n",
                g.nfconf.num_tids, g.nfconf.num_ports, g.nfconf.num_links);

        /* calculate masks */
        tid_len = (int)ceil(log2((double)g.nfconf.num_tids));
        pl_len = (int)ceil(log2((double)(g.nfconf.num_links * 
                                         g.nfconf.num_ports)));

        mask = pick_mask(tid_len);
        g.nfconf.tid_mask = mask << ZF_FILTER_TID_OFF;
        mask = pick_mask(pl_len);
        g.nfconf.pl_off = ZF_FILTER_TID_OFF + tid_len;
        g.nfconf.pl_mask = mask << g.nfconf.pl_off;
        dprint(LOG_DBG, "TID_MASK: %08x  PL_MASK: %08x\n", g.nfconf.tid_mask,
               g.nfconf.pl_mask);
}


int main(int argc, char **argv)
{
        int err, offset;

        if (argc < 1)
                usage();

        /* default values that need memory allocation (so that we can free() them.
         * These could be static as well, but since the user may override them, 
         * freeing the memory becomes a bit harder. We would have to make sure 
         * we are not freeing from the stack.
         */
        psmconf.iface = strdup("eth0");
        g.opt.nf2_dev = strdup("nf2c0");
        g.cleanup.iface = psmconf.iface;

        offset = parse_args_opts(argc, argv);

        if (!SIMULATE) {
                g.nf2 = nflib_open(g.opt.nf2_dev);
                if (!g.nf2) {
                        dprint(LOG_ERR, "Unable to connect to a NetFPGA device "
                               "on interface: %s\n", g.opt.nf2_dev);
                        global_error_exit("");
                }
        }

        basic_setup();

        err = parse_args(argc, argv, offset, &psmconf);
        if (err) {
                global_error_exit("Processing commands failed");
        }
                                        
        /* now start executing... */
        execute(&psmconf);
        global_exit();
        return 0;
}


